@@ -1,21 +1,120 @@
 from abc import ABCMeta
 from functools import partial
-from inspect import ismodule
+from inspect import isclass, ismodule
 from pathlib import Path
-from typing import Callable, List, Union
+from typing import Any, Callable, List, Optional, Union
 
 from .config import RegistryConfig
-from .exceptions import CannotRegisterPythonBuiltInError, ModuleAliasError
+from .exceptions import (
+    CannotDeriveNameError,
+    CannotRegisterPythonBuiltInError,
+    InvalidNameError,
+    KeyCollisionError,
+    ModuleAliasError,
+)
+
+
+class _Registry(dict):
+    """Unified container object for __registry__."""
+
+    def __init__(self, config: RegistryConfig, name: Optional[str] = None):
+        super().__init__()
+        self.config = config
+        self.name = name
+
+        # These will be populated later
+        self.cls: Any = None
+
+    def register(
+        self,
+        obj: Any,
+        name: Union[str, None] = None,
+        aliases: Union[str, None, List[str]] = None,
+        root: bool = False,
+    ):
+        """Register an object to a registry, subject to configuration.
+
+        Parameters
+        ----------
+        obj: object
+            object to store and attempt to auto-derive name from.
+        name: str
+            If provided, register ``obj`` to this name; overrides checks.
+            If not provided, name will be auto-derived from ``obj`` via ``format``.
+        aliases: Union[str, None, List[str]]
+            If provided, also register ``obj`` to these strings.
+            Not subject to configuration rules.
+        root: bool
+            Set to ``True`` when calling initial ``__register__``.
+            Force register to immediate parent(s).
+        """
+        # Derive/Validate Name
+        if name is None:
+            try:
+                name = str(obj.__name__)
+            except AttributeError:
+                raise CannotDeriveNameError(
+                    f"Cannot derive name from a bare {type(obj)}."
+                )
+            name = self.config.format(name)
+        elif "." in name or "/" in name:
+            raise InvalidNameError(f'Name "{name}" cannot contain "." or "/".')
+
+        if not self.config.overwrite and name in self:
+            raise KeyCollisionError(f'"{name}" already registered to {self}')
+
+        # Validate aliases and massage it into a list.
+        if aliases is None:
+            aliases = []
+        elif isinstance(aliases, str):
+            aliases = [aliases]
+
+        for alias in aliases:
+            if "." in alias or "/" in alias:
+                raise InvalidNameError(f'Alias "{alias}" cannot contain "." or "/".')
+            if not self.config.overwrite and alias in self:
+                raise KeyCollisionError(f'"{alias}" already registered to {self}')
+
+        # Check if should register self
+        if obj == self.cls:
+            if self.config.register_self:
+                self[name] = obj
+        else:
+            self[name] = obj
+
+        # Register to parents if one of the following conditions are met:
+        #     1. This is the root ``__recursive__`` call.
+        #     2. Both this.recursive is True, and parent.recursive is True.
+        if (root or self.config.recursive) and self.cls is not None:
+            for parent_cls in self.cls.__bases__:
+                try:
+                    parent_registry = parent_cls.__registry__
+                except AttributeError:
+                    # Not a Registry object
+                    continue
+
+                if parent_cls is Registry:
+                    # Only register to Registry if this is a root call.
+                    if root:
+                        parent_registry.register(obj, name=name, aliases=aliases)
+                elif root or parent_registry.config.recursive:
+                    parent_registry.register(obj, name=name, aliases=aliases)
+
+        # Register aliases
+        for alias in aliases:
+            if not self.config.overwrite and alias in self:
+                raise KeyCollisionError(f'"{alias}" already registered to {self}')
+
+            self[alias] = obj
 
 
 class _DictMixin:
     """Dict-like methods for a registry-based class."""
 
-    __registry__: dict
-    __registry_config__: RegistryConfig
+    __registry__: _Registry
 
     def __getitem__(self, key: str):
-        return self.__registry_config__.getitem(self.__registry__, key)
+        return self.__registry__.config.getitem(self.__registry__, key)
 
     def __iter__(self):
         for val in self.__registry__:
@@ -26,7 +125,7 @@ class _DictMixin:
 
     def __contains__(self, key: str):
         try:
-            self.__registry_config__.getitem(self.__registry__, key)
+            self.__registry__.config.getitem(self.__registry__, key)
         except KeyError:
             return False
         return True
@@ -78,9 +177,7 @@ class _RedirectMethod(object):
 
 
 class RegistryMeta(ABCMeta, _DictMixin):
-    __registry__: dict
-    __registry_config__: RegistryConfig
-    __registry_name__: str
+    __registry__: _Registry
 
     def __new__(
         mcls,
@@ -107,36 +204,37 @@ class RegistryMeta(ABCMeta, _DictMixin):
         # Manipulate namespace instead of modifying attributes after calling __new__ so
         # that hooks like __init_subclass__ have appropriately set registry attributes.
         # Each subclass gets its own registry.
-        namespace["__registry__"] = {}
         try:
             Registry
         except NameError:
             # Should only happen the very first time that
             # Registry is being defined.
+            namespace["__registry__"] = _Registry(RegistryConfig(**config))
             cls = super().__new__(mcls, cls_name, bases, namespace)
-            cls.__registry_config__ = RegistryConfig(**config)
             return cls
 
         # Copy the nearest parent config, then update it with new params
         for parent_cls in bases:
             try:
-                namespace["__registry_config__"] = parent_cls.__registry_config__.copy()
+                registry_config = parent_cls.__registry__.config.copy()
                 break
             except AttributeError:
                 pass
+        else:
+            raise Exception("Should never happen.")  # pragma: no cover
 
-        # Set __registry_name__ before updating __registry_config__, since a classes own name is
+        # Derive registry name before updating registry config, since a classes own name is
         # subject to it's parents configuration, not its own.
         if name is None:
-            namespace["__registry_name__"] = namespace["__registry_config__"].format(
-                cls_name
-            )
+            registry_name = registry_config.format(cls_name)
         else:
-            namespace["__registry_name__"] = name
+            registry_name = name
 
-        namespace["__registry_config__"].update(config)
+        registry_config.update(config)
 
-        if namespace["__registry_config__"].redirect:
+        namespace["__registry__"] = _Registry(registry_config, name=registry_name)
+
+        if namespace["__registry__"].config.redirect:
             for key in [
                 "__getitem__",
                 "__iter__",
@@ -156,42 +254,18 @@ class RegistryMeta(ABCMeta, _DictMixin):
         # We cannot defer class creation any further.
         # This will call hooks like __init_subclass__
         cls = super().__new__(mcls, cls_name, bases, namespace)
+        cls.__registry__.cls = cls
 
         if skip:
             return cls
 
-        # Register direct subclasses of Register to Register
-        if cls in Registry.__subclasses__() and cls_name != "RegistryDecorator":
-            Registry.__registry_config__.register(
-                Registry.__registry__, cls, name=cls.__registry_name__
-            )
-
-        # otherwise, register it in own registry and all parent registries.
-        for parent_cls in cls.mro():
-            if parent_cls == Registry:
-                continue
-
-            if (
-                not cls.__registry_config__.recursive
-                and cls not in parent_cls.__subclasses__()
-            ):
-                continue
-
-            try:
-                config = parent_cls.__registry_config__  # type: ignore
-            except AttributeError:
-                # Not a Registry object
-                continue
-
-            if not config.register_self and parent_cls == cls:
-                continue
-
-            config.register(
-                parent_cls.__registry__,
-                cls,
-                name=cls.__registry_name__,
-                aliases=aliases,
-            )  # type: ignore
+        # Register to parent(s)
+        cls.__registry__.register(
+            cls,
+            name=cls.__registry__.name,
+            aliases=aliases,
+            root=True,  # Always register to direct parents
+        )
 
         return cls
 
@@ -230,8 +304,7 @@ class RegistryDecorator(Registry, _DictMixin):
         """Create a Registry for decorating."""
         # overwrite the registry data so its independent
         # of the Registry object.
-        self.__registry__ = {}
-        self.__registry_config__ = RegistryConfig(**config)
+        self.__registry__ = _Registry(RegistryConfig(**config))
 
         if objs is None:
             objs = []
@@ -249,7 +322,7 @@ class RegistryDecorator(Registry, _DictMixin):
         name=None,
         aliases: Union[str, None, List[str]] = None,
     ):
-        config = self.__registry_config__
+        config = self.__registry__.config
 
         if obj is None:
             # Was called @my_registry(**config_params)
@@ -257,7 +330,7 @@ class RegistryDecorator(Registry, _DictMixin):
             return partial(self.__call__, name=name, aliases=aliases)
 
         if not ismodule(obj):
-            config.register(self.__registry__, obj, name=name, aliases=aliases)
+            self.__registry__.register(obj, name=name, aliases=aliases)
             return obj
 
         if aliases:
